-file("/usr/local/lib/erlang/lib/parsetools-2.1.1/include/leexinc.hrl", 0).
%% The source of this file is part of leex distribution, as such it
%% has the same Copyright as the other files in the leex
%% distribution. The Copyright is defined in the accompanying file
%% COPYRIGHT. However, the resultant scanner generated by leex is the
%% property of the creator of the scanner and is not covered by that
%% Copyright.

-module(gpb_scan).

-export([string/1,string/2,token/2,token/3,tokens/2,tokens/3]).
-export([format_error/1]).

%% User code. This is placed here to allow extra attributes.
-file("src/gpb_scan.xrl", 120).

-include_lib("eunit/include/eunit.hrl").

%% Eliminate a dialyzer warning like below:
%% /usr/lib/erlang/lib/parsetools-2.1.1/include/leexinc.hrl:268:
%%  Function yyrev/2 will never be called
-dialyzer({nowarn_function, yyrev/2}).

string_value(S) -> % S is with start+end quote
    %% Strip quotes.
    string_val_2(lists:sublist(S, 2, length(S) - 2)).

string_val_2([$\\|Cs]) -> string_escape(Cs);
string_val_2([C|Cs])   -> [C|string_val_2(Cs)];
string_val_2([])       -> [].

-define(is_octal_char(C), $0 =< C, C =< $7).

string_escape("x"++Rest) ->
    {HexChars, Rest2} = collect(fun is_hex_char/1, Rest),
    [hex_to_integer(HexChars) | string_val_2(Rest2)];
string_escape([Oct|Rest]) when ?is_octal_char(Oct) ->
    {OctChars, Rest2} = collect(fun is_oct_char/1, [Oct|Rest]),
    [oct_to_integer(OctChars) | string_val_2(Rest2)];
string_escape([C|Rest]) ->
    [escape_char(C) | string_val_2(Rest)].

escape_char($a) -> 7;                           %\a = BEL
escape_char($b) -> $\b;                         %\b = BS
escape_char($f) -> $\f;                         %\f = FF
escape_char($n) -> $\n;                         %\n = LF
escape_char($r) -> $\r;                         %\r = CR
escape_char($t) -> $\t;                         %\t = TAB
escape_char($v) -> $\v;                         %\v = VT
escape_char(C)  -> C.

collect(Pred, Str) -> lists:splitwith(Pred, Str).

is_oct_char(C) -> $0 =< C andalso C =< $7.
is_hex_char(C) -> ($0 =< C andalso C =< $9)
                      orelse ($a =< C andalso C =< $f)
                      orelse ($A =< C andalso C =< $F).

oct_to_integer(Str) -> erlang:list_to_integer(Str, 8).
hex_to_integer(Str) -> erlang:list_to_integer(Str, 16).

octstr_to_integer(Str) -> oct_to_integer(Str).

hexstr_to_integer("0x"++H)  -> hex_to_integer(H);
hexstr_to_integer("0X"++H)  -> hex_to_integer(H);
hexstr_to_integer("+0x"++H) -> hex_to_integer(H);
hexstr_to_integer("+0X"++H) -> hex_to_integer(H);
hexstr_to_integer("-0x"++H) -> -hex_to_integer(H);
hexstr_to_integer("-0X"++H) -> -hex_to_integer(H).

str_to_float_1(S)        -> list_to_float(S).

str_to_float_2("."++_=S) -> str_to_float_1("0"++S);
str_to_float_2("+."++T)  -> str_to_float_1("0."++T);
str_to_float_2("-."++T)  -> str_to_float_1("-0."++T).

str_to_float_3(S) -> %% No decimals after `.' Possibly e+-<n> following `.'
    {UpToDot, "."++Rest} = collect(fun isnt_dot/1, S),
    str_to_float_1(UpToDot++"."++"0"++Rest).

str_to_float_4(S) -> %% Integer preceding e+-<n>
    {UpToDot, Rest} = collect(fun isnt_exp_e/1, S),
    str_to_float_1(UpToDot++"."++"0"++Rest).

isnt_dot(C) -> C /= $. .

isnt_exp_e($e) -> false;
isnt_exp_e($E) -> false;
isnt_exp_e(_)  -> true.

-file("/usr/local/lib/erlang/lib/parsetools-2.1.1/include/leexinc.hrl", 14).

format_error({illegal,S}) -> ["illegal characters ",io_lib:write_string(S)];
format_error({user,S}) -> S.

string(String) -> string(String, 1).

string(String, Line) -> string(String, Line, String, []).

%% string(InChars, Line, TokenChars, Tokens) ->
%% {ok,Tokens,Line} | {error,ErrorInfo,Line}.
%% Note the line number going into yystate, L0, is line of token
%% start while line number returned is line of token end. We want line
%% of token start.

string([], L, [], Ts) ->                     % No partial tokens!
    {ok,yyrev(Ts),L};
string(Ics0, L0, Tcs, Ts) ->
    case yystate(yystate(), Ics0, L0, 0, reject, 0) of
        {A,Alen,Ics1,L1} ->                  % Accepting end state
            string_cont(Ics1, L1, yyaction(A, Alen, Tcs, L0), Ts);
        {A,Alen,Ics1,L1,_S1} ->              % Accepting transistion state
            string_cont(Ics1, L1, yyaction(A, Alen, Tcs, L0), Ts);
        {reject,_Alen,Tlen,_Ics1,L1,_S1} ->  % After a non-accepting state
            {error,{L0,?MODULE,{illegal,yypre(Tcs, Tlen+1)}},L1};
        {A,Alen,_Tlen,_Ics1,_L1,_S1} ->
            string_cont(yysuf(Tcs, Alen), L0, yyaction(A, Alen, Tcs, L0), Ts)
    end.

%% string_cont(RestChars, Line, Token, Tokens)
%% Test for and remove the end token wrapper. Push back characters
%% are prepended to RestChars.

-dialyzer({nowarn_function, string_cont/4}).

string_cont(Rest, Line, {token,T}, Ts) ->
    string(Rest, Line, Rest, [T|Ts]);
string_cont(Rest, Line, {token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    string(NewRest, Line, NewRest, [T|Ts]);
string_cont(Rest, Line, {end_token,T}, Ts) ->
    string(Rest, Line, Rest, [T|Ts]);
string_cont(Rest, Line, {end_token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    string(NewRest, Line, NewRest, [T|Ts]);
string_cont(Rest, Line, skip_token, Ts) ->
    string(Rest, Line, Rest, Ts);
string_cont(Rest, Line, {skip_token,Push}, Ts) ->
    NewRest = Push ++ Rest,
    string(NewRest, Line, NewRest, Ts);
string_cont(_Rest, Line, {error,S}, _Ts) ->
    {error,{Line,?MODULE,{user,S}},Line}.

%% token(Continuation, Chars) ->
%% token(Continuation, Chars, Line) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.
%% Must be careful when re-entering to append the latest characters to the
%% after characters in an accept. The continuation is:
%% {token,State,CurrLine,TokenChars,TokenLen,TokenLine,AccAction,AccLen}

token(Cont, Chars) -> token(Cont, Chars, 1).

token([], Chars, Line) ->
    token(yystate(), Chars, Line, Chars, 0, Line, reject, 0);
token({token,State,Line,Tcs,Tlen,Tline,Action,Alen}, Chars, _) ->
    token(State, Chars, Line, Tcs ++ Chars, Tlen, Tline, Action, Alen).

%% token(State, InChars, Line, TokenChars, TokenLen, TokenLine,
%% AcceptAction, AcceptLen) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.
%% The argument order is chosen to be more efficient.

token(S0, Ics0, L0, Tcs, Tlen0, Tline, A0, Alen0) ->
    case yystate(S0, Ics0, L0, Tlen0, A0, Alen0) of
        %% Accepting end state, we have a token.
        {A1,Alen1,Ics1,L1} ->
            token_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline));
        %% Accepting transition state, can take more chars.
        {A1,Alen1,[],L1,S1} ->                  % Need more chars to check
            {more,{token,S1,L1,Tcs,Alen1,Tline,A1,Alen1}};
        {A1,Alen1,Ics1,L1,_S1} ->               % Take what we got
            token_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline));
        %% After a non-accepting state, maybe reach accept state later.
        {A1,Alen1,Tlen1,[],L1,S1} ->            % Need more chars to check
            {more,{token,S1,L1,Tcs,Tlen1,Tline,A1,Alen1}};
        {reject,_Alen1,Tlen1,eof,L1,_S1} ->     % No token match
            %% Check for partial token which is error.
            Ret = if Tlen1 > 0 -> {error,{Tline,?MODULE,
                                          %% Skip eof tail in Tcs.
                                          {illegal,yypre(Tcs, Tlen1)}},L1};
                     true -> {eof,L1}
                  end,
            {done,Ret,eof};
        {reject,_Alen1,Tlen1,Ics1,L1,_S1} ->    % No token match
            Error = {Tline,?MODULE,{illegal,yypre(Tcs, Tlen1+1)}},
            {done,{error,Error,L1},Ics1};
        {A1,Alen1,_Tlen1,_Ics1,_L1,_S1} ->       % Use last accept match
            token_cont(yysuf(Tcs, Alen1), L0, yyaction(A1, Alen1, Tcs, Tline))
    end.

%% token_cont(RestChars, Line, Token)
%% If we have a token or error then return done, else if we have a
%% skip_token then continue.

-dialyzer({nowarn_function, token_cont/3}).

token_cont(Rest, Line, {token,T}) ->
    {done,{ok,T,Line},Rest};
token_cont(Rest, Line, {token,T,Push}) ->
    NewRest = Push ++ Rest,
    {done,{ok,T,Line},NewRest};
token_cont(Rest, Line, {end_token,T}) ->
    {done,{ok,T,Line},Rest};
token_cont(Rest, Line, {end_token,T,Push}) ->
    NewRest = Push ++ Rest,
    {done,{ok,T,Line},NewRest};
token_cont(Rest, Line, skip_token) ->
    token(yystate(), Rest, Line, Rest, 0, Line, reject, 0);
token_cont(Rest, Line, {skip_token,Push}) ->
    NewRest = Push ++ Rest,
    token(yystate(), NewRest, Line, NewRest, 0, Line, reject, 0);
token_cont(Rest, Line, {error,S}) ->
    {done,{error,{Line,?MODULE,{user,S}},Line},Rest}.

%% tokens(Continuation, Chars, Line) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.
%% Must be careful when re-entering to append the latest characters to the
%% after characters in an accept. The continuation is:
%% {tokens,State,CurrLine,TokenChars,TokenLen,TokenLine,Tokens,AccAction,AccLen}
%% {skip_tokens,State,CurrLine,TokenChars,TokenLen,TokenLine,Error,AccAction,AccLen}

tokens(Cont, Chars) -> tokens(Cont, Chars, 1).

tokens([], Chars, Line) ->
    tokens(yystate(), Chars, Line, Chars, 0, Line, [], reject, 0);
tokens({tokens,State,Line,Tcs,Tlen,Tline,Ts,Action,Alen}, Chars, _) ->
    tokens(State, Chars, Line, Tcs ++ Chars, Tlen, Tline, Ts, Action, Alen);
tokens({skip_tokens,State,Line,Tcs,Tlen,Tline,Error,Action,Alen}, Chars, _) ->
    skip_tokens(State, Chars, Line, Tcs ++ Chars, Tlen, Tline, Error, Action, Alen).

%% tokens(State, InChars, Line, TokenChars, TokenLen, TokenLine, Tokens,
%% AcceptAction, AcceptLen) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.

tokens(S0, Ics0, L0, Tcs, Tlen0, Tline, Ts, A0, Alen0) ->
    case yystate(S0, Ics0, L0, Tlen0, A0, Alen0) of
        %% Accepting end state, we have a token.
        {A1,Alen1,Ics1,L1} ->
            tokens_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Ts);
        %% Accepting transition state, can take more chars.
        {A1,Alen1,[],L1,S1} ->                  % Need more chars to check
            {more,{tokens,S1,L1,Tcs,Alen1,Tline,Ts,A1,Alen1}};
        {A1,Alen1,Ics1,L1,_S1} ->               % Take what we got
            tokens_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Ts);
        %% After a non-accepting state, maybe reach accept state later.
        {A1,Alen1,Tlen1,[],L1,S1} ->            % Need more chars to check
            {more,{tokens,S1,L1,Tcs,Tlen1,Tline,Ts,A1,Alen1}};
        {reject,_Alen1,Tlen1,eof,L1,_S1} ->     % No token match
            %% Check for partial token which is error, no need to skip here.
            Ret = if Tlen1 > 0 -> {error,{Tline,?MODULE,
                                          %% Skip eof tail in Tcs.
                                          {illegal,yypre(Tcs, Tlen1)}},L1};
                     Ts == [] -> {eof,L1};
                     true -> {ok,yyrev(Ts),L1}
                  end,
            {done,Ret,eof};
        {reject,_Alen1,Tlen1,_Ics1,L1,_S1} ->
            %% Skip rest of tokens.
            Error = {L1,?MODULE,{illegal,yypre(Tcs, Tlen1+1)}},
            skip_tokens(yysuf(Tcs, Tlen1+1), L1, Error);
        {A1,Alen1,_Tlen1,_Ics1,_L1,_S1} ->
            Token = yyaction(A1, Alen1, Tcs, Tline),
            tokens_cont(yysuf(Tcs, Alen1), L0, Token, Ts)
    end.

%% tokens_cont(RestChars, Line, Token, Tokens)
%% If we have an end_token or error then return done, else if we have
%% a token then save it and continue, else if we have a skip_token
%% just continue.

-dialyzer({nowarn_function, tokens_cont/4}).

tokens_cont(Rest, Line, {token,T}, Ts) ->
    tokens(yystate(), Rest, Line, Rest, 0, Line, [T|Ts], reject, 0);
tokens_cont(Rest, Line, {token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    tokens(yystate(), NewRest, Line, NewRest, 0, Line, [T|Ts], reject, 0);
tokens_cont(Rest, Line, {end_token,T}, Ts) ->
    {done,{ok,yyrev(Ts, [T]),Line},Rest};
tokens_cont(Rest, Line, {end_token,T,Push}, Ts) ->
    NewRest = Push ++ Rest,
    {done,{ok,yyrev(Ts, [T]),Line},NewRest};
tokens_cont(Rest, Line, skip_token, Ts) ->
    tokens(yystate(), Rest, Line, Rest, 0, Line, Ts, reject, 0);
tokens_cont(Rest, Line, {skip_token,Push}, Ts) ->
    NewRest = Push ++ Rest,
    tokens(yystate(), NewRest, Line, NewRest, 0, Line, Ts, reject, 0);
tokens_cont(Rest, Line, {error,S}, _Ts) ->
    skip_tokens(Rest, Line, {Line,?MODULE,{user,S}}).

%%skip_tokens(InChars, Line, Error) -> {done,{error,Error,Line},Ics}.
%% Skip tokens until an end token, junk everything and return the error.

skip_tokens(Ics, Line, Error) ->
    skip_tokens(yystate(), Ics, Line, Ics, 0, Line, Error, reject, 0).

%% skip_tokens(State, InChars, Line, TokenChars, TokenLen, TokenLine, Tokens,
%% AcceptAction, AcceptLen) ->
%% {more,Continuation} | {done,ReturnVal,RestChars}.

skip_tokens(S0, Ics0, L0, Tcs, Tlen0, Tline, Error, A0, Alen0) ->
    case yystate(S0, Ics0, L0, Tlen0, A0, Alen0) of
        {A1,Alen1,Ics1,L1} ->                  % Accepting end state
            skip_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Error);
        {A1,Alen1,[],L1,S1} ->                 % After an accepting state
            {more,{skip_tokens,S1,L1,Tcs,Alen1,Tline,Error,A1,Alen1}};
        {A1,Alen1,Ics1,L1,_S1} ->
            skip_cont(Ics1, L1, yyaction(A1, Alen1, Tcs, Tline), Error);
        {A1,Alen1,Tlen1,[],L1,S1} ->           % After a non-accepting state
            {more,{skip_tokens,S1,L1,Tcs,Tlen1,Tline,Error,A1,Alen1}};
        {reject,_Alen1,_Tlen1,eof,L1,_S1} ->
            {done,{error,Error,L1},eof};
        {reject,_Alen1,Tlen1,_Ics1,L1,_S1} ->
            skip_tokens(yysuf(Tcs, Tlen1+1), L1, Error);
        {A1,Alen1,_Tlen1,_Ics1,L1,_S1} ->
            Token = yyaction(A1, Alen1, Tcs, Tline),
            skip_cont(yysuf(Tcs, Alen1), L1, Token, Error)
    end.

%% skip_cont(RestChars, Line, Token, Error)
%% Skip tokens until we have an end_token or error then return done
%% with the original rror.

-dialyzer({nowarn_function, skip_cont/4}).

skip_cont(Rest, Line, {token,_T}, Error) ->
    skip_tokens(yystate(), Rest, Line, Rest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {token,_T,Push}, Error) ->
    NewRest = Push ++ Rest,
    skip_tokens(yystate(), NewRest, Line, NewRest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {end_token,_T}, Error) ->
    {done,{error,Error,Line},Rest};
skip_cont(Rest, Line, {end_token,_T,Push}, Error) ->
    NewRest = Push ++ Rest,
    {done,{error,Error,Line},NewRest};
skip_cont(Rest, Line, skip_token, Error) ->
    skip_tokens(yystate(), Rest, Line, Rest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {skip_token,Push}, Error) ->
    NewRest = Push ++ Rest,
    skip_tokens(yystate(), NewRest, Line, NewRest, 0, Line, Error, reject, 0);
skip_cont(Rest, Line, {error,_S}, Error) ->
    skip_tokens(yystate(), Rest, Line, Rest, 0, Line, Error, reject, 0).

yyrev(List) -> lists:reverse(List).
yyrev(List, Tail) -> lists:reverse(List, Tail).
yypre(List, N) -> lists:sublist(List, N).
yysuf(List, N) -> lists:nthtail(N, List).

%% yystate() -> InitialState.
%% yystate(State, InChars, Line, CurrTokLen, AcceptAction, AcceptLen) ->
%% {Action, AcceptLen, RestChars, Line} |
%% {Action, AcceptLen, RestChars, Line, State} |
%% {reject, AcceptLen, CurrTokLen, RestChars, Line, State} |
%% {Action, AcceptLen, CurrTokLen, RestChars, Line, State}.
%% Generated state transition functions. The non-accepting end state
%% return signal either an unrecognised character or end of current
%% input.

-file("src/gpb_scan.erl", 358).
yystate() -> 224.

yystate(227, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(227, [50|Ics], Line, Tlen, _, _) ->
    yystate(225, Ics, Line, Tlen+1, 48, Tlen);
yystate(227, [48|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(227, [49|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(227, [C|Ics], Line, Tlen, _, _) when C >= 51, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(227, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(227, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(227, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,227};
yystate(226, [10|Ics], Line, Tlen, _, _) ->
    yystate(226, Ics, Line+1, Tlen+1, 64, Tlen);
yystate(226, [C|Ics], Line, Tlen, _, _) when C >= 0, C =< 9 ->
    yystate(226, Ics, Line, Tlen+1, 64, Tlen);
yystate(226, [C|Ics], Line, Tlen, _, _) when C >= 11, C =< 32 ->
    yystate(226, Ics, Line, Tlen+1, 64, Tlen);
yystate(226, Ics, Line, Tlen, _, _) ->
    {64,Tlen,Ics,Line,226};
yystate(225, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 6, Tlen);
yystate(225, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 6, Tlen);
yystate(225, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 6, Tlen);
yystate(225, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 6, Tlen);
yystate(225, Ics, Line, Tlen, _, _) ->
    {6,Tlen,Ics,Line,225};
yystate(224, [125|Ics], Line, Tlen, Action, Alen) ->
    yystate(220, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [123|Ics], Line, Tlen, Action, Alen) ->
    yystate(216, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [117|Ics], Line, Tlen, Action, Alen) ->
    yystate(212, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [116|Ics], Line, Tlen, Action, Alen) ->
    yystate(176, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [115|Ics], Line, Tlen, Action, Alen) ->
    yystate(156, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [114|Ics], Line, Tlen, Action, Alen) ->
    yystate(24, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [113|Ics], Line, Tlen, Action, Alen) ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [112|Ics], Line, Tlen, Action, Alen) ->
    yystate(83, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [111|Ics], Line, Tlen, Action, Alen) ->
    yystate(119, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [110|Ics], Line, Tlen, Action, Alen) ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [109|Ics], Line, Tlen, Action, Alen) ->
    yystate(167, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [105|Ics], Line, Tlen, Action, Alen) ->
    yystate(207, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [103|Ics], Line, Tlen, Action, Alen) ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [104|Ics], Line, Tlen, Action, Alen) ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [102|Ics], Line, Tlen, Action, Alen) ->
    yystate(201, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [101|Ics], Line, Tlen, Action, Alen) ->
    yystate(133, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [100|Ics], Line, Tlen, Action, Alen) ->
    yystate(77, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [99|Ics], Line, Tlen, Action, Alen) ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [98|Ics], Line, Tlen, Action, Alen) ->
    yystate(2, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [97|Ics], Line, Tlen, Action, Alen) ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [95|Ics], Line, Tlen, Action, Alen) ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [93|Ics], Line, Tlen, Action, Alen) ->
    yystate(34, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [91|Ics], Line, Tlen, Action, Alen) ->
    yystate(38, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [62|Ics], Line, Tlen, Action, Alen) ->
    yystate(42, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [61|Ics], Line, Tlen, Action, Alen) ->
    yystate(46, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [60|Ics], Line, Tlen, Action, Alen) ->
    yystate(50, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [59|Ics], Line, Tlen, Action, Alen) ->
    yystate(54, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [48|Ics], Line, Tlen, Action, Alen) ->
    yystate(106, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [47|Ics], Line, Tlen, Action, Alen) ->
    yystate(122, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [46|Ics], Line, Tlen, Action, Alen) ->
    yystate(146, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [45|Ics], Line, Tlen, Action, Alen) ->
    yystate(174, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [44|Ics], Line, Tlen, Action, Alen) ->
    yystate(170, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [43|Ics], Line, Tlen, Action, Alen) ->
    yystate(174, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [41|Ics], Line, Tlen, Action, Alen) ->
    yystate(178, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [40|Ics], Line, Tlen, Action, Alen) ->
    yystate(182, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [39|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [34|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(226, Ics, Line+1, Tlen+1, Action, Alen);
yystate(224, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(226, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 32 ->
    yystate(226, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [C|Ics], Line, Tlen, Action, Alen) when C >= 49, C =< 57 ->
    yystate(58, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [C|Ics], Line, Tlen, Action, Alen) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [C|Ics], Line, Tlen, Action, Alen) when C >= 106, C =< 108 ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, [C|Ics], Line, Tlen, Action, Alen) when C >= 118, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, Action, Alen);
yystate(224, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,224};
yystate(223, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 7, Tlen);
yystate(223, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 7, Tlen);
yystate(223, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 7, Tlen);
yystate(223, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 7, Tlen);
yystate(223, Ics, Line, Tlen, _, _) ->
    {7,Tlen,Ics,Line,223};
yystate(222, [92|Ics], Line, Tlen, Action, Alen) ->
    yystate(210, Ics, Line, Tlen+1, Action, Alen);
yystate(222, [34|Ics], Line, Tlen, Action, Alen) ->
    yystate(218, Ics, Line, Tlen+1, Action, Alen);
yystate(222, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(222, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 33 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(222, [C|Ics], Line, Tlen, Action, Alen) when C >= 35, C =< 91 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(222, [C|Ics], Line, Tlen, Action, Alen) when C >= 93 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(222, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,222};
yystate(221, [112|Ics], Line, Tlen, _, _) ->
    yystate(217, Ics, Line, Tlen+1, 48, Tlen);
yystate(221, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(221, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(221, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(221, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 111 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(221, [C|Ics], Line, Tlen, _, _) when C >= 113, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(221, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,221};
yystate(220, Ics, Line, Tlen, _, _) ->
    {52,Tlen,Ics,Line};
yystate(219, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(219, [52|Ics], Line, Tlen, _, _) ->
    yystate(223, Ics, Line, Tlen+1, 48, Tlen);
yystate(219, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 51 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(219, [C|Ics], Line, Tlen, _, _) when C >= 53, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(219, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(219, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(219, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,219};
yystate(218, Ics, Line, Tlen, _, _) ->
    {39,Tlen,Ics,Line};
yystate(217, [111|Ics], Line, Tlen, _, _) ->
    yystate(213, Ics, Line, Tlen+1, 48, Tlen);
yystate(217, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(217, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(217, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(217, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(217, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(217, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,217};
yystate(216, Ics, Line, Tlen, _, _) ->
    {51,Tlen,Ics,Line};
yystate(215, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [54|Ics], Line, Tlen, _, _) ->
    yystate(219, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [52|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [53|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [51|Ics], Line, Tlen, _, _) ->
    yystate(227, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 50 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [C|Ics], Line, Tlen, _, _) when C >= 55, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(215, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,215};
yystate(214, [92|Ics], Line, Tlen, _, _) ->
    yystate(210, Ics, Line, Tlen+1, 39, Tlen);
yystate(214, [34|Ics], Line, Tlen, _, _) ->
    yystate(218, Ics, Line, Tlen+1, 39, Tlen);
yystate(214, [C|Ics], Line, Tlen, _, _) when C >= 0, C =< 9 ->
    yystate(222, Ics, Line, Tlen+1, 39, Tlen);
yystate(214, [C|Ics], Line, Tlen, _, _) when C >= 11, C =< 33 ->
    yystate(222, Ics, Line, Tlen+1, 39, Tlen);
yystate(214, [C|Ics], Line, Tlen, _, _) when C >= 35, C =< 91 ->
    yystate(222, Ics, Line, Tlen+1, 39, Tlen);
yystate(214, [C|Ics], Line, Tlen, _, _) when C >= 93 ->
    yystate(222, Ics, Line, Tlen+1, 39, Tlen);
yystate(214, Ics, Line, Tlen, _, _) ->
    {39,Tlen,Ics,Line,214};
yystate(213, [114|Ics], Line, Tlen, _, _) ->
    yystate(209, Ics, Line, Tlen+1, 48, Tlen);
yystate(213, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(213, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(213, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(213, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(213, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(213, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,213};
yystate(212, [105|Ics], Line, Tlen, _, _) ->
    yystate(208, Ics, Line, Tlen+1, 48, Tlen);
yystate(212, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(212, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(212, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(212, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(212, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(212, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,212};
yystate(211, [116|Ics], Line, Tlen, _, _) ->
    yystate(215, Ics, Line, Tlen+1, 48, Tlen);
yystate(211, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(211, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(211, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(211, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(211, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(211, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,211};
yystate(210, [120|Ics], Line, Tlen, Action, Alen) ->
    yystate(206, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [119|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [118|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [117|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [116|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [115|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [114|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [110|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [102|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [97|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [98|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [92|Ics], Line, Tlen, Action, Alen) ->
    yystate(210, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [63|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [39|Ics], Line, Tlen, Action, Alen) ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [34|Ics], Line, Tlen, Action, Alen) ->
    yystate(214, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 33 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 35, C =< 38 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 40, C =< 47 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 55 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 56, C =< 62 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 64, C =< 91 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 93, C =< 96 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 99, C =< 101 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 103, C =< 109 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 111, C =< 113 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, [C|Ics], Line, Tlen, Action, Alen) when C >= 121 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(210, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,210};
yystate(209, [116|Ics], Line, Tlen, _, _) ->
    yystate(205, Ics, Line, Tlen+1, 48, Tlen);
yystate(209, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(209, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(209, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(209, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(209, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(209, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,209};
yystate(208, [110|Ics], Line, Tlen, _, _) ->
    yystate(204, Ics, Line, Tlen+1, 48, Tlen);
yystate(208, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(208, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(208, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(208, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(208, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(208, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,208};
yystate(207, [110|Ics], Line, Tlen, _, _) ->
    yystate(211, Ics, Line, Tlen+1, 48, Tlen);
yystate(207, [109|Ics], Line, Tlen, _, _) ->
    yystate(221, Ics, Line, Tlen+1, 48, Tlen);
yystate(207, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(207, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(207, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(207, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 108 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(207, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(207, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,207};
yystate(206, [92|Ics], Line, Tlen, Action, Alen) ->
    yystate(210, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [34|Ics], Line, Tlen, Action, Alen) ->
    yystate(218, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 33 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 35, C =< 47 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 58, C =< 64 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 65, C =< 70 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 71, C =< 91 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 93, C =< 96 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 97, C =< 102 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, [C|Ics], Line, Tlen, Action, Alen) when C >= 103 ->
    yystate(222, Ics, Line, Tlen+1, Action, Alen);
yystate(206, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,206};
yystate(205, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 24, Tlen);
yystate(205, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 24, Tlen);
yystate(205, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 24, Tlen);
yystate(205, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 24, Tlen);
yystate(205, Ics, Line, Tlen, _, _) ->
    {24,Tlen,Ics,Line,205};
yystate(204, [116|Ics], Line, Tlen, _, _) ->
    yystate(200, Ics, Line, Tlen+1, 48, Tlen);
yystate(204, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(204, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(204, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(204, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(204, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(204, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,204};
yystate(203, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 19, Tlen);
yystate(203, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 19, Tlen);
yystate(203, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 19, Tlen);
yystate(203, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 19, Tlen);
yystate(203, Ics, Line, Tlen, _, _) ->
    {19,Tlen,Ics,Line,203};
yystate(202, [92|Ics], Line, Tlen, Action, Alen) ->
    yystate(190, Ics, Line, Tlen+1, Action, Alen);
yystate(202, [39|Ics], Line, Tlen, Action, Alen) ->
    yystate(198, Ics, Line, Tlen+1, Action, Alen);
yystate(202, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(202, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 38 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(202, [C|Ics], Line, Tlen, Action, Alen) when C >= 40, C =< 91 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(202, [C|Ics], Line, Tlen, Action, Alen) when C >= 93 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(202, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,202};
yystate(201, [108|Ics], Line, Tlen, _, _) ->
    yystate(197, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [106|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [107|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [105|Ics], Line, Tlen, _, _) ->
    yystate(181, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [97|Ics], Line, Tlen, _, _) ->
    yystate(149, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(201, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,201};
yystate(200, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [54|Ics], Line, Tlen, _, _) ->
    yystate(196, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [52|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [53|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [51|Ics], Line, Tlen, _, _) ->
    yystate(188, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 50 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [C|Ics], Line, Tlen, _, _) when C >= 55, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(200, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,200};
yystate(199, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 31, Tlen);
yystate(199, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 31, Tlen);
yystate(199, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 31, Tlen);
yystate(199, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 31, Tlen);
yystate(199, Ics, Line, Tlen, _, _) ->
    {31,Tlen,Ics,Line,199};
yystate(198, Ics, Line, Tlen, _, _) ->
    {40,Tlen,Ics,Line};
yystate(197, [111|Ics], Line, Tlen, _, _) ->
    yystate(193, Ics, Line, Tlen+1, 48, Tlen);
yystate(197, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(197, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(197, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(197, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(197, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(197, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,197};
yystate(196, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(196, [52|Ics], Line, Tlen, _, _) ->
    yystate(192, Ics, Line, Tlen+1, 48, Tlen);
yystate(196, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 51 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(196, [C|Ics], Line, Tlen, _, _) when C >= 53, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(196, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(196, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(196, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,196};
yystate(195, [121|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [122|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [120|Ics], Line, Tlen, _, _) ->
    yystate(199, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [112|Ics], Line, Tlen, _, _) ->
    yystate(203, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 111 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, [C|Ics], Line, Tlen, _, _) when C >= 113, C =< 119 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(195, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,195};
yystate(194, [92|Ics], Line, Tlen, _, _) ->
    yystate(190, Ics, Line, Tlen+1, 40, Tlen);
yystate(194, [39|Ics], Line, Tlen, _, _) ->
    yystate(198, Ics, Line, Tlen+1, 40, Tlen);
yystate(194, [C|Ics], Line, Tlen, _, _) when C >= 0, C =< 9 ->
    yystate(202, Ics, Line, Tlen+1, 40, Tlen);
yystate(194, [C|Ics], Line, Tlen, _, _) when C >= 11, C =< 38 ->
    yystate(202, Ics, Line, Tlen+1, 40, Tlen);
yystate(194, [C|Ics], Line, Tlen, _, _) when C >= 40, C =< 91 ->
    yystate(202, Ics, Line, Tlen+1, 40, Tlen);
yystate(194, [C|Ics], Line, Tlen, _, _) when C >= 93 ->
    yystate(202, Ics, Line, Tlen+1, 40, Tlen);
yystate(194, Ics, Line, Tlen, _, _) ->
    {40,Tlen,Ics,Line,194};
yystate(193, [97|Ics], Line, Tlen, _, _) ->
    yystate(189, Ics, Line, Tlen+1, 48, Tlen);
yystate(193, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(193, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(193, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(193, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(193, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,193};
yystate(192, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 9, Tlen);
yystate(192, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 9, Tlen);
yystate(192, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 9, Tlen);
yystate(192, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 9, Tlen);
yystate(192, Ics, Line, Tlen, _, _) ->
    {9,Tlen,Ics,Line,192};
yystate(191, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 21, Tlen);
yystate(191, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 21, Tlen);
yystate(191, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 21, Tlen);
yystate(191, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 21, Tlen);
yystate(191, Ics, Line, Tlen, _, _) ->
    {21,Tlen,Ics,Line,191};
yystate(190, [120|Ics], Line, Tlen, Action, Alen) ->
    yystate(186, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [119|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [118|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [117|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [116|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [115|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [114|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [110|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [102|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [97|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [98|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [92|Ics], Line, Tlen, Action, Alen) ->
    yystate(190, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [63|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [39|Ics], Line, Tlen, Action, Alen) ->
    yystate(194, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [34|Ics], Line, Tlen, Action, Alen) ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 33 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 35, C =< 38 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 40, C =< 47 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 55 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 56, C =< 62 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 64, C =< 91 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 93, C =< 96 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 99, C =< 101 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 103, C =< 109 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 111, C =< 113 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, [C|Ics], Line, Tlen, Action, Alen) when C >= 121 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(190, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,190};
yystate(189, [116|Ics], Line, Tlen, _, _) ->
    yystate(185, Ics, Line, Tlen+1, 48, Tlen);
yystate(189, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(189, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(189, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(189, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(189, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(189, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,189};
yystate(188, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(188, [50|Ics], Line, Tlen, _, _) ->
    yystate(184, Ics, Line, Tlen+1, 48, Tlen);
yystate(188, [48|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(188, [49|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(188, [C|Ics], Line, Tlen, _, _) when C >= 51, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(188, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(188, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(188, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,188};
yystate(187, [101|Ics], Line, Tlen, _, _) ->
    yystate(191, Ics, Line, Tlen+1, 48, Tlen);
yystate(187, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(187, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(187, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(187, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(187, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(187, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,187};
yystate(186, [92|Ics], Line, Tlen, Action, Alen) ->
    yystate(190, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [39|Ics], Line, Tlen, Action, Alen) ->
    yystate(198, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 38 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 40, C =< 47 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 58, C =< 64 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 65, C =< 70 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 71, C =< 91 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 93, C =< 96 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 97, C =< 102 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, [C|Ics], Line, Tlen, Action, Alen) when C >= 103 ->
    yystate(202, Ics, Line, Tlen+1, Action, Alen);
yystate(186, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,186};
yystate(185, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 5, Tlen);
yystate(185, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 5, Tlen);
yystate(185, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 5, Tlen);
yystate(185, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 5, Tlen);
yystate(185, Ics, Line, Tlen, _, _) ->
    {5,Tlen,Ics,Line,185};
yystate(184, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 8, Tlen);
yystate(184, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 8, Tlen);
yystate(184, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 8, Tlen);
yystate(184, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 8, Tlen);
yystate(184, Ics, Line, Tlen, _, _) ->
    {8,Tlen,Ics,Line,184};
yystate(183, [103|Ics], Line, Tlen, _, _) ->
    yystate(187, Ics, Line, Tlen+1, 48, Tlen);
yystate(183, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(183, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(183, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(183, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 102 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(183, [C|Ics], Line, Tlen, _, _) when C >= 104, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(183, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,183};
yystate(182, Ics, Line, Tlen, _, _) ->
    {55,Tlen,Ics,Line};
yystate(181, [121|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(181, [122|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(181, [120|Ics], Line, Tlen, _, _) ->
    yystate(177, Ics, Line, Tlen+1, 48, Tlen);
yystate(181, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(181, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(181, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(181, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 119 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(181, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,181};
yystate(180, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(180, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(180, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(180, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(180, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,180};
yystate(179, [97|Ics], Line, Tlen, _, _) ->
    yystate(183, Ics, Line, Tlen+1, 48, Tlen);
yystate(179, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(179, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(179, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(179, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(179, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,179};
yystate(178, Ics, Line, Tlen, _, _) ->
    {56,Tlen,Ics,Line};
yystate(177, [101|Ics], Line, Tlen, _, _) ->
    yystate(173, Ics, Line, Tlen+1, 48, Tlen);
yystate(177, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(177, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(177, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(177, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(177, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(177, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,177};
yystate(176, [114|Ics], Line, Tlen, _, _) ->
    yystate(172, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [112|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [113|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [111|Ics], Line, Tlen, _, _) ->
    yystate(160, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(176, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,176};
yystate(175, [115|Ics], Line, Tlen, _, _) ->
    yystate(179, Ics, Line, Tlen+1, 48, Tlen);
yystate(175, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(175, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(175, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(175, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(175, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(175, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,175};
yystate(174, [48|Ics], Line, Tlen, Action, Alen) ->
    yystate(106, Ics, Line, Tlen+1, Action, Alen);
yystate(174, [46|Ics], Line, Tlen, Action, Alen) ->
    yystate(166, Ics, Line, Tlen+1, Action, Alen);
yystate(174, [C|Ics], Line, Tlen, Action, Alen) when C >= 49, C =< 57 ->
    yystate(58, Ics, Line, Tlen+1, Action, Alen);
yystate(174, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,174};
yystate(173, [100|Ics], Line, Tlen, _, _) ->
    yystate(169, Ics, Line, Tlen+1, 48, Tlen);
yystate(173, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(173, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(173, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(173, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(173, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(173, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,173};
yystate(172, [117|Ics], Line, Tlen, _, _) ->
    yystate(168, Ics, Line, Tlen+1, 48, Tlen);
yystate(172, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(172, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(172, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(172, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 116 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(172, [C|Ics], Line, Tlen, _, _) when C >= 118, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(172, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,172};
yystate(171, [115|Ics], Line, Tlen, _, _) ->
    yystate(175, Ics, Line, Tlen+1, 48, Tlen);
yystate(171, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(171, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(171, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(171, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(171, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(171, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,171};
yystate(170, Ics, Line, Tlen, _, _) ->
    {58,Tlen,Ics,Line};
yystate(169, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [54|Ics], Line, Tlen, _, _) ->
    yystate(165, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [52|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [53|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [51|Ics], Line, Tlen, _, _) ->
    yystate(157, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 50 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [C|Ics], Line, Tlen, _, _) when C >= 55, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(169, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,169};
yystate(168, [101|Ics], Line, Tlen, _, _) ->
    yystate(164, Ics, Line, Tlen+1, 48, Tlen);
yystate(168, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(168, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(168, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(168, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(168, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(168, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,168};
yystate(167, [101|Ics], Line, Tlen, _, _) ->
    yystate(171, Ics, Line, Tlen+1, 48, Tlen);
yystate(167, [97|Ics], Line, Tlen, _, _) ->
    yystate(195, Ics, Line, Tlen+1, 48, Tlen);
yystate(167, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(167, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(167, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(167, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(167, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(167, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,167};
yystate(166, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(150, Ics, Line, Tlen+1, Action, Alen);
yystate(166, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,166};
yystate(165, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(165, [52|Ics], Line, Tlen, _, _) ->
    yystate(161, Ics, Line, Tlen+1, 48, Tlen);
yystate(165, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 51 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(165, [C|Ics], Line, Tlen, _, _) when C >= 53, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(165, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(165, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(165, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,165};
yystate(164, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 36, Tlen);
yystate(164, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 36, Tlen);
yystate(164, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 36, Tlen);
yystate(164, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 36, Tlen);
yystate(164, Ics, Line, Tlen, _, _) ->
    {36,Tlen,Ics,Line,164};
yystate(163, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 3, Tlen);
yystate(163, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 3, Tlen);
yystate(163, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 3, Tlen);
yystate(163, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 3, Tlen);
yystate(163, Ics, Line, Tlen, _, _) ->
    {3,Tlen,Ics,Line,163};
yystate(162, [45|Ics], Line, Tlen, Action, Alen) ->
    yystate(158, Ics, Line, Tlen+1, Action, Alen);
yystate(162, [43|Ics], Line, Tlen, Action, Alen) ->
    yystate(158, Ics, Line, Tlen+1, Action, Alen);
yystate(162, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(154, Ics, Line, Tlen+1, Action, Alen);
yystate(162, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,162};
yystate(161, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 13, Tlen);
yystate(161, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 13, Tlen);
yystate(161, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 13, Tlen);
yystate(161, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 13, Tlen);
yystate(161, Ics, Line, Tlen, _, _) ->
    {13,Tlen,Ics,Line,161};
yystate(160, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 30, Tlen);
yystate(160, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 30, Tlen);
yystate(160, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 30, Tlen);
yystate(160, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 30, Tlen);
yystate(160, Ics, Line, Tlen, _, _) ->
    {30,Tlen,Ics,Line,160};
yystate(159, [102|Ics], Line, Tlen, _, _) ->
    yystate(163, Ics, Line, Tlen+1, 48, Tlen);
yystate(159, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(159, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(159, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(159, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 101 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(159, [C|Ics], Line, Tlen, _, _) when C >= 103, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(159, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,159};
yystate(158, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(154, Ics, Line, Tlen+1, Action, Alen);
yystate(158, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,158};
yystate(157, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(157, [50|Ics], Line, Tlen, _, _) ->
    yystate(153, Ics, Line, Tlen+1, 48, Tlen);
yystate(157, [48|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(157, [49|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(157, [C|Ics], Line, Tlen, _, _) when C >= 51, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(157, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(157, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(157, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,157};
yystate(156, [122|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [121|Ics], Line, Tlen, _, _) ->
    yystate(152, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [116|Ics], Line, Tlen, _, _) ->
    yystate(132, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [105|Ics], Line, Tlen, _, _) ->
    yystate(112, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [103|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [104|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [102|Ics], Line, Tlen, _, _) ->
    yystate(84, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [101|Ics], Line, Tlen, _, _) ->
    yystate(48, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 120 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(156, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,156};
yystate(155, [111|Ics], Line, Tlen, _, _) ->
    yystate(159, Ics, Line, Tlen+1, 48, Tlen);
yystate(155, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(155, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(155, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(155, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(155, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(155, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,155};
yystate(154, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(154, Ics, Line, Tlen+1, 42, Tlen);
yystate(154, Ics, Line, Tlen, _, _) ->
    {42,Tlen,Ics,Line,154};
yystate(153, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 12, Tlen);
yystate(153, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 12, Tlen);
yystate(153, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 12, Tlen);
yystate(153, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 12, Tlen);
yystate(153, Ics, Line, Tlen, _, _) ->
    {12,Tlen,Ics,Line,153};
yystate(152, [110|Ics], Line, Tlen, _, _) ->
    yystate(148, Ics, Line, Tlen+1, 48, Tlen);
yystate(152, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(152, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(152, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(152, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(152, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(152, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,152};
yystate(151, [101|Ics], Line, Tlen, _, _) ->
    yystate(155, Ics, Line, Tlen+1, 48, Tlen);
yystate(151, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(151, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(151, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(151, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(151, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(151, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,151};
yystate(150, [101|Ics], Line, Tlen, _, _) ->
    yystate(162, Ics, Line, Tlen+1, 42, Tlen);
yystate(150, [69|Ics], Line, Tlen, _, _) ->
    yystate(162, Ics, Line, Tlen+1, 42, Tlen);
yystate(150, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(150, Ics, Line, Tlen+1, 42, Tlen);
yystate(150, Ics, Line, Tlen, _, _) ->
    {42,Tlen,Ics,Line,150};
yystate(149, [108|Ics], Line, Tlen, _, _) ->
    yystate(145, Ics, Line, Tlen+1, 48, Tlen);
yystate(149, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(149, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(149, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(149, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(149, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(149, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,149};
yystate(148, [116|Ics], Line, Tlen, _, _) ->
    yystate(144, Ics, Line, Tlen+1, 48, Tlen);
yystate(148, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(148, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(148, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(148, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(148, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(148, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,148};
yystate(147, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 1, Tlen);
yystate(147, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 1, Tlen);
yystate(147, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 1, Tlen);
yystate(147, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 1, Tlen);
yystate(147, Ics, Line, Tlen, _, _) ->
    {1,Tlen,Ics,Line,147};
yystate(146, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(150, Ics, Line, Tlen+1, 49, Tlen);
yystate(146, Ics, Line, Tlen, _, _) ->
    {49,Tlen,Ics,Line,146};
yystate(145, [115|Ics], Line, Tlen, _, _) ->
    yystate(141, Ics, Line, Tlen+1, 48, Tlen);
yystate(145, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(145, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(145, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(145, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(145, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(145, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,145};
yystate(144, [97|Ics], Line, Tlen, _, _) ->
    yystate(140, Ics, Line, Tlen+1, 48, Tlen);
yystate(144, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(144, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(144, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(144, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(144, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,144};
yystate(143, [108|Ics], Line, Tlen, _, _) ->
    yystate(147, Ics, Line, Tlen+1, 48, Tlen);
yystate(143, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(143, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(143, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(143, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(143, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(143, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,143};
yystate(142, [42|Ics], Line, Tlen, Action, Alen) ->
    yystate(134, Ics, Line, Tlen+1, Action, Alen);
yystate(142, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(142, Ics, Line+1, Tlen+1, Action, Alen);
yystate(142, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(142, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 41 ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(142, [C|Ics], Line, Tlen, Action, Alen) when C >= 43 ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(142, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,142};
yystate(141, [101|Ics], Line, Tlen, _, _) ->
    yystate(137, Ics, Line, Tlen+1, 48, Tlen);
yystate(141, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(141, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(141, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,141};
yystate(140, [121|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(140, [122|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(140, [120|Ics], Line, Tlen, _, _) ->
    yystate(136, Ics, Line, Tlen+1, 48, Tlen);
yystate(140, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(140, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(140, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(140, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 119 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(140, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,140};
yystate(139, [97|Ics], Line, Tlen, _, _) ->
    yystate(143, Ics, Line, Tlen+1, 23, Tlen);
yystate(139, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 23, Tlen);
yystate(139, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 23, Tlen);
yystate(139, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 23, Tlen);
yystate(139, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 23, Tlen);
yystate(139, Ics, Line, Tlen, _, _) ->
    {23,Tlen,Ics,Line,139};
yystate(138, Ics, Line, Tlen, _, _) ->
    {63,Tlen,Ics,Line};
yystate(137, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 37, Tlen);
yystate(137, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 37, Tlen);
yystate(137, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 37, Tlen);
yystate(137, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 37, Tlen);
yystate(137, Ics, Line, Tlen, _, _) ->
    {37,Tlen,Ics,Line,137};
yystate(136, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 38, Tlen);
yystate(136, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 38, Tlen);
yystate(136, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 38, Tlen);
yystate(136, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 38, Tlen);
yystate(136, Ics, Line, Tlen, _, _) ->
    {38,Tlen,Ics,Line,136};
yystate(135, [110|Ics], Line, Tlen, _, _) ->
    yystate(139, Ics, Line, Tlen+1, 48, Tlen);
yystate(135, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(135, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(135, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,135};
yystate(134, [47|Ics], Line, Tlen, Action, Alen) ->
    yystate(138, Ics, Line, Tlen+1, Action, Alen);
yystate(134, [42|Ics], Line, Tlen, Action, Alen) ->
    yystate(134, Ics, Line, Tlen+1, Action, Alen);
yystate(134, [10|Ics], Line, Tlen, Action, Alen) ->
    yystate(142, Ics, Line+1, Tlen+1, Action, Alen);
yystate(134, [C|Ics], Line, Tlen, Action, Alen) when C >= 0, C =< 9 ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(134, [C|Ics], Line, Tlen, Action, Alen) when C >= 11, C =< 41 ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(134, [C|Ics], Line, Tlen, Action, Alen) when C >= 43, C =< 46 ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(134, [C|Ics], Line, Tlen, Action, Alen) when C >= 48 ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(134, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,134};
yystate(133, [121|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [122|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [120|Ics], Line, Tlen, _, _) ->
    yystate(129, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [110|Ics], Line, Tlen, _, _) ->
    yystate(89, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 119 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(133, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,133};
yystate(132, [114|Ics], Line, Tlen, _, _) ->
    yystate(128, Ics, Line, Tlen+1, 48, Tlen);
yystate(132, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(132, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(132, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(132, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(132, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(132, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,132};
yystate(131, [111|Ics], Line, Tlen, _, _) ->
    yystate(135, Ics, Line, Tlen+1, 48, Tlen);
yystate(131, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(131, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(131, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,131};
yystate(130, [10|Ics], Line, Tlen, _, _) ->
    yystate(126, Ics, Line+1, Tlen+1, 62, Tlen);
yystate(130, [C|Ics], Line, Tlen, _, _) when C >= 0, C =< 9 ->
    yystate(130, Ics, Line, Tlen+1, 62, Tlen);
yystate(130, [C|Ics], Line, Tlen, _, _) when C >= 11 ->
    yystate(130, Ics, Line, Tlen+1, 62, Tlen);
yystate(130, Ics, Line, Tlen, _, _) ->
    {62,Tlen,Ics,Line,130};
yystate(129, [116|Ics], Line, Tlen, _, _) ->
    yystate(125, Ics, Line, Tlen+1, 48, Tlen);
yystate(129, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(129, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(129, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,129};
yystate(128, [105|Ics], Line, Tlen, _, _) ->
    yystate(124, Ics, Line, Tlen+1, 48, Tlen);
yystate(128, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(128, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(128, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(128, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(128, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(128, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,128};
yystate(127, [105|Ics], Line, Tlen, _, _) ->
    yystate(131, Ics, Line, Tlen+1, 48, Tlen);
yystate(127, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(127, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(127, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(127, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(127, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(127, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,127};
yystate(126, Ics, Line, Tlen, _, _) ->
    {61,Tlen,Ics,Line};
yystate(125, [101|Ics], Line, Tlen, _, _) ->
    yystate(121, Ics, Line, Tlen+1, 48, Tlen);
yystate(125, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(125, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(125, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,125};
yystate(124, [110|Ics], Line, Tlen, _, _) ->
    yystate(120, Ics, Line, Tlen+1, 48, Tlen);
yystate(124, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(124, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(124, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(124, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(124, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(124, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,124};
yystate(123, [116|Ics], Line, Tlen, _, _) ->
    yystate(127, Ics, Line, Tlen+1, 48, Tlen);
yystate(123, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(123, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(123, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,123};
yystate(122, [47|Ics], Line, Tlen, Action, Alen) ->
    yystate(130, Ics, Line, Tlen+1, Action, Alen);
yystate(122, [42|Ics], Line, Tlen, Action, Alen) ->
    yystate(142, Ics, Line, Tlen+1, Action, Alen);
yystate(122, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,122};
yystate(121, [110|Ics], Line, Tlen, _, _) ->
    yystate(117, Ics, Line, Tlen+1, 48, Tlen);
yystate(121, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(121, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(121, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,121};
yystate(120, [103|Ics], Line, Tlen, _, _) ->
    yystate(116, Ics, Line, Tlen+1, 48, Tlen);
yystate(120, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(120, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(120, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(120, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 102 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(120, [C|Ics], Line, Tlen, _, _) when C >= 104, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(120, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,120};
yystate(119, [112|Ics], Line, Tlen, _, _) ->
    yystate(123, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, [111|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, [110|Ics], Line, Tlen, _, _) ->
    yystate(151, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, [C|Ics], Line, Tlen, _, _) when C >= 113, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(119, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,119};
yystate(118, [101|Ics], Line, Tlen, _, _) ->
    yystate(70, Ics, Line, Tlen+1, 46, Tlen);
yystate(118, [69|Ics], Line, Tlen, _, _) ->
    yystate(70, Ics, Line, Tlen+1, 46, Tlen);
yystate(118, [46|Ics], Line, Tlen, _, _) ->
    yystate(74, Ics, Line, Tlen+1, 46, Tlen);
yystate(118, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(118, Ics, Line, Tlen+1, 46, Tlen);
yystate(118, Ics, Line, Tlen, _, _) ->
    {46,Tlen,Ics,Line,118};
yystate(117, [115|Ics], Line, Tlen, _, _) ->
    yystate(113, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, [100|Ics], Line, Tlen, _, _) ->
    yystate(93, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 114 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(117, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,117};
yystate(116, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 17, Tlen);
yystate(116, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 17, Tlen);
yystate(116, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 17, Tlen);
yystate(116, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 17, Tlen);
yystate(116, Ics, Line, Tlen, _, _) ->
    {17,Tlen,Ics,Line,116};
yystate(115, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 20, Tlen);
yystate(115, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 20, Tlen);
yystate(115, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 20, Tlen);
yystate(115, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 20, Tlen);
yystate(115, Ics, Line, Tlen, _, _) ->
    {20,Tlen,Ics,Line,115};
yystate(114, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(114, Ics, Line, Tlen+1, 45, Tlen);
yystate(114, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 70 ->
    yystate(114, Ics, Line, Tlen+1, 45, Tlen);
yystate(114, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 102 ->
    yystate(114, Ics, Line, Tlen+1, 45, Tlen);
yystate(114, Ics, Line, Tlen, _, _) ->
    {45,Tlen,Ics,Line,114};
yystate(113, [105|Ics], Line, Tlen, _, _) ->
    yystate(109, Ics, Line, Tlen+1, 48, Tlen);
yystate(113, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(113, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(113, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(113, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(113, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(113, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,113};
yystate(112, [110|Ics], Line, Tlen, _, _) ->
    yystate(108, Ics, Line, Tlen+1, 48, Tlen);
yystate(112, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(112, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(112, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(112, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(112, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(112, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,112};
yystate(111, [101|Ics], Line, Tlen, _, _) ->
    yystate(115, Ics, Line, Tlen+1, 48, Tlen);
yystate(111, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(111, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(111, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(111, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(111, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(111, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,111};
yystate(110, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(114, Ics, Line, Tlen+1, Action, Alen);
yystate(110, [C|Ics], Line, Tlen, Action, Alen) when C >= 65, C =< 70 ->
    yystate(114, Ics, Line, Tlen+1, Action, Alen);
yystate(110, [C|Ics], Line, Tlen, Action, Alen) when C >= 97, C =< 102 ->
    yystate(114, Ics, Line, Tlen+1, Action, Alen);
yystate(110, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,110};
yystate(109, [111|Ics], Line, Tlen, _, _) ->
    yystate(105, Ics, Line, Tlen+1, 48, Tlen);
yystate(109, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(109, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(109, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(109, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(109, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(109, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,109};
yystate(108, [116|Ics], Line, Tlen, _, _) ->
    yystate(104, Ics, Line, Tlen+1, 48, Tlen);
yystate(108, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(108, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(108, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(108, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(108, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(108, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,108};
yystate(107, [103|Ics], Line, Tlen, _, _) ->
    yystate(111, Ics, Line, Tlen+1, 48, Tlen);
yystate(107, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 102 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(107, [C|Ics], Line, Tlen, _, _) when C >= 104, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(107, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,107};
yystate(106, [120|Ics], Line, Tlen, _, _) ->
    yystate(110, Ics, Line, Tlen+1, 47, Tlen);
yystate(106, [101|Ics], Line, Tlen, _, _) ->
    yystate(70, Ics, Line, Tlen+1, 47, Tlen);
yystate(106, [88|Ics], Line, Tlen, _, _) ->
    yystate(110, Ics, Line, Tlen+1, 47, Tlen);
yystate(106, [69|Ics], Line, Tlen, _, _) ->
    yystate(70, Ics, Line, Tlen+1, 47, Tlen);
yystate(106, [46|Ics], Line, Tlen, _, _) ->
    yystate(74, Ics, Line, Tlen+1, 47, Tlen);
yystate(106, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(118, Ics, Line, Tlen+1, 47, Tlen);
yystate(106, Ics, Line, Tlen, _, _) ->
    {47,Tlen,Ics,Line,106};
yystate(105, [110|Ics], Line, Tlen, _, _) ->
    yystate(101, Ics, Line, Tlen+1, 48, Tlen);
yystate(105, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(105, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(105, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,105};
yystate(104, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [54|Ics], Line, Tlen, _, _) ->
    yystate(100, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [52|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [53|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [51|Ics], Line, Tlen, _, _) ->
    yystate(92, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 50 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [C|Ics], Line, Tlen, _, _) when C >= 55, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(104, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,104};
yystate(103, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 26, Tlen);
yystate(103, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 26, Tlen);
yystate(103, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 26, Tlen);
yystate(103, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 26, Tlen);
yystate(103, Ics, Line, Tlen, _, _) ->
    {26,Tlen,Ics,Line,103};
yystate(102, [45|Ics], Line, Tlen, Action, Alen) ->
    yystate(98, Ics, Line, Tlen+1, Action, Alen);
yystate(102, [43|Ics], Line, Tlen, Action, Alen) ->
    yystate(98, Ics, Line, Tlen+1, Action, Alen);
yystate(102, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(102, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,102};
yystate(101, [115|Ics], Line, Tlen, _, _) ->
    yystate(97, Ics, Line, Tlen+1, 48, Tlen);
yystate(101, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(101, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(101, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,101};
yystate(100, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(100, [52|Ics], Line, Tlen, _, _) ->
    yystate(96, Ics, Line, Tlen+1, 48, Tlen);
yystate(100, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 51 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(100, [C|Ics], Line, Tlen, _, _) when C >= 53, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(100, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(100, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(100, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,100};
yystate(99, [100|Ics], Line, Tlen, _, _) ->
    yystate(103, Ics, Line, Tlen+1, 48, Tlen);
yystate(99, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(99, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(99, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,99};
yystate(98, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, Action, Alen);
yystate(98, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,98};
yystate(97, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 28, Tlen);
yystate(97, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 28, Tlen);
yystate(97, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 28, Tlen);
yystate(97, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 28, Tlen);
yystate(97, Ics, Line, Tlen, _, _) ->
    {28,Tlen,Ics,Line,97};
yystate(96, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 11, Tlen);
yystate(96, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 11, Tlen);
yystate(96, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 11, Tlen);
yystate(96, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 11, Tlen);
yystate(96, Ics, Line, Tlen, _, _) ->
    {11,Tlen,Ics,Line,96};
yystate(95, [101|Ics], Line, Tlen, _, _) ->
    yystate(99, Ics, Line, Tlen+1, 48, Tlen);
yystate(95, [97|Ics], Line, Tlen, _, _) ->
    yystate(107, Ics, Line, Tlen+1, 48, Tlen);
yystate(95, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(95, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(95, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,95};
yystate(94, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(94, Ics, Line, Tlen+1, 41, Tlen);
yystate(94, Ics, Line, Tlen, _, _) ->
    {41,Tlen,Ics,Line,94};
yystate(93, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 29, Tlen);
yystate(93, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 29, Tlen);
yystate(93, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 29, Tlen);
yystate(93, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 29, Tlen);
yystate(93, Ics, Line, Tlen, _, _) ->
    {29,Tlen,Ics,Line,93};
yystate(92, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(92, [50|Ics], Line, Tlen, _, _) ->
    yystate(88, Ics, Line, Tlen+1, 48, Tlen);
yystate(92, [48|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(92, [49|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(92, [C|Ics], Line, Tlen, _, _) when C >= 51, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(92, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(92, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(92, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,92};
yystate(91, [107|Ics], Line, Tlen, _, _) ->
    yystate(95, Ics, Line, Tlen+1, 48, Tlen);
yystate(91, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(91, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(91, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(91, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 106 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(91, [C|Ics], Line, Tlen, _, _) when C >= 108, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(91, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,91};
yystate(90, [101|Ics], Line, Tlen, _, _) ->
    yystate(102, Ics, Line, Tlen+1, 41, Tlen);
yystate(90, [69|Ics], Line, Tlen, _, _) ->
    yystate(102, Ics, Line, Tlen+1, 41, Tlen);
yystate(90, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(90, Ics, Line, Tlen+1, 41, Tlen);
yystate(90, Ics, Line, Tlen, _, _) ->
    {41,Tlen,Ics,Line,90};
yystate(89, [117|Ics], Line, Tlen, _, _) ->
    yystate(85, Ics, Line, Tlen+1, 48, Tlen);
yystate(89, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(89, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(89, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(89, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 116 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(89, [C|Ics], Line, Tlen, _, _) when C >= 118, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(89, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,89};
yystate(88, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 10, Tlen);
yystate(88, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 10, Tlen);
yystate(88, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 10, Tlen);
yystate(88, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 10, Tlen);
yystate(88, Ics, Line, Tlen, _, _) ->
    {10,Tlen,Ics,Line,88};
yystate(87, [99|Ics], Line, Tlen, _, _) ->
    yystate(91, Ics, Line, Tlen+1, 48, Tlen);
yystate(87, [97|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(87, [98|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(87, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(87, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(87, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(87, [C|Ics], Line, Tlen, _, _) when C >= 100, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(87, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,87};
yystate(86, [45|Ics], Line, Tlen, Action, Alen) ->
    yystate(82, Ics, Line, Tlen+1, Action, Alen);
yystate(86, [43|Ics], Line, Tlen, Action, Alen) ->
    yystate(82, Ics, Line, Tlen+1, Action, Alen);
yystate(86, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(78, Ics, Line, Tlen+1, Action, Alen);
yystate(86, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,86};
yystate(85, [109|Ics], Line, Tlen, _, _) ->
    yystate(81, Ics, Line, Tlen+1, 48, Tlen);
yystate(85, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 108 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(85, [C|Ics], Line, Tlen, _, _) when C >= 110, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(85, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,85};
yystate(84, [105|Ics], Line, Tlen, _, _) ->
    yystate(80, Ics, Line, Tlen+1, 48, Tlen);
yystate(84, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(84, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(84, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(84, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(84, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(84, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,84};
yystate(83, [97|Ics], Line, Tlen, _, _) ->
    yystate(87, Ics, Line, Tlen+1, 48, Tlen);
yystate(83, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(83, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(83, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(83, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(83, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,83};
yystate(82, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(78, Ics, Line, Tlen+1, Action, Alen);
yystate(82, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,82};
yystate(81, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 22, Tlen);
yystate(81, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 22, Tlen);
yystate(81, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 22, Tlen);
yystate(81, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 22, Tlen);
yystate(81, Ics, Line, Tlen, _, _) ->
    {22,Tlen,Ics,Line,81};
yystate(80, [121|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(80, [122|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(80, [120|Ics], Line, Tlen, _, _) ->
    yystate(76, Ics, Line, Tlen+1, 48, Tlen);
yystate(80, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(80, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(80, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(80, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 119 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(80, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,80};
yystate(79, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 2, Tlen);
yystate(79, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 2, Tlen);
yystate(79, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 2, Tlen);
yystate(79, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 2, Tlen);
yystate(79, Ics, Line, Tlen, _, _) ->
    {2,Tlen,Ics,Line,79};
yystate(78, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(78, Ics, Line, Tlen+1, 43, Tlen);
yystate(78, Ics, Line, Tlen, _, _) ->
    {43,Tlen,Ics,Line,78};
yystate(77, [111|Ics], Line, Tlen, _, _) ->
    yystate(73, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, [101|Ics], Line, Tlen, _, _) ->
    yystate(53, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(77, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,77};
yystate(76, [101|Ics], Line, Tlen, _, _) ->
    yystate(72, Ics, Line, Tlen+1, 48, Tlen);
yystate(76, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(76, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(76, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(76, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(76, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(76, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,76};
yystate(75, [100|Ics], Line, Tlen, _, _) ->
    yystate(79, Ics, Line, Tlen+1, 48, Tlen);
yystate(75, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(75, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(75, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(75, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(75, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(75, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,75};
yystate(74, [101|Ics], Line, Tlen, _, _) ->
    yystate(86, Ics, Line, Tlen+1, 43, Tlen);
yystate(74, [69|Ics], Line, Tlen, _, _) ->
    yystate(86, Ics, Line, Tlen+1, 43, Tlen);
yystate(74, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(90, Ics, Line, Tlen+1, 43, Tlen);
yystate(74, Ics, Line, Tlen, _, _) ->
    {43,Tlen,Ics,Line,74};
yystate(73, [117|Ics], Line, Tlen, _, _) ->
    yystate(69, Ics, Line, Tlen+1, 48, Tlen);
yystate(73, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 116 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(73, [C|Ics], Line, Tlen, _, _) when C >= 118, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(73, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,73};
yystate(72, [100|Ics], Line, Tlen, _, _) ->
    yystate(68, Ics, Line, Tlen+1, 48, Tlen);
yystate(72, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(72, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(72, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(72, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(72, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(72, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,72};
yystate(71, [101|Ics], Line, Tlen, _, _) ->
    yystate(75, Ics, Line, Tlen+1, 48, Tlen);
yystate(71, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(71, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(71, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,71};
yystate(70, [45|Ics], Line, Tlen, Action, Alen) ->
    yystate(66, Ics, Line, Tlen+1, Action, Alen);
yystate(70, [43|Ics], Line, Tlen, Action, Alen) ->
    yystate(66, Ics, Line, Tlen+1, Action, Alen);
yystate(70, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(62, Ics, Line, Tlen+1, Action, Alen);
yystate(70, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,70};
yystate(69, [98|Ics], Line, Tlen, _, _) ->
    yystate(65, Ics, Line, Tlen+1, 48, Tlen);
yystate(69, [97|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(69, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(69, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(69, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(69, [C|Ics], Line, Tlen, _, _) when C >= 99, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(69, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,69};
yystate(68, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [54|Ics], Line, Tlen, _, _) ->
    yystate(64, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [52|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [53|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [51|Ics], Line, Tlen, _, _) ->
    yystate(56, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 50 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [C|Ics], Line, Tlen, _, _) when C >= 55, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(68, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,68};
yystate(67, [116|Ics], Line, Tlen, _, _) ->
    yystate(71, Ics, Line, Tlen+1, 48, Tlen);
yystate(67, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(67, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(67, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,67};
yystate(66, [C|Ics], Line, Tlen, Action, Alen) when C >= 48, C =< 57 ->
    yystate(62, Ics, Line, Tlen+1, Action, Alen);
yystate(66, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,66};
yystate(65, [108|Ics], Line, Tlen, _, _) ->
    yystate(61, Ics, Line, Tlen+1, 48, Tlen);
yystate(65, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(65, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(65, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,65};
yystate(64, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(64, [52|Ics], Line, Tlen, _, _) ->
    yystate(60, Ics, Line, Tlen+1, 48, Tlen);
yystate(64, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 51 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(64, [C|Ics], Line, Tlen, _, _) when C >= 53, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(64, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(64, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(64, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,64};
yystate(63, [97|Ics], Line, Tlen, _, _) ->
    yystate(67, Ics, Line, Tlen+1, 48, Tlen);
yystate(63, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(63, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(63, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(63, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(63, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,63};
yystate(62, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(62, Ics, Line, Tlen+1, 44, Tlen);
yystate(62, Ics, Line, Tlen, _, _) ->
    {44,Tlen,Ics,Line,62};
yystate(61, [101|Ics], Line, Tlen, _, _) ->
    yystate(57, Ics, Line, Tlen+1, 48, Tlen);
yystate(61, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(61, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(61, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(61, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(61, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(61, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,61};
yystate(60, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 15, Tlen);
yystate(60, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 15, Tlen);
yystate(60, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 15, Tlen);
yystate(60, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 15, Tlen);
yystate(60, Ics, Line, Tlen, _, _) ->
    {15,Tlen,Ics,Line,60};
yystate(59, [101|Ics], Line, Tlen, _, _) ->
    yystate(63, Ics, Line, Tlen+1, 48, Tlen);
yystate(59, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(59, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(59, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(59, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(59, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(59, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,59};
yystate(58, [101|Ics], Line, Tlen, _, _) ->
    yystate(70, Ics, Line, Tlen+1, 47, Tlen);
yystate(58, [69|Ics], Line, Tlen, _, _) ->
    yystate(70, Ics, Line, Tlen+1, 47, Tlen);
yystate(58, [46|Ics], Line, Tlen, _, _) ->
    yystate(74, Ics, Line, Tlen+1, 47, Tlen);
yystate(58, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(58, Ics, Line, Tlen+1, 47, Tlen);
yystate(58, Ics, Line, Tlen, _, _) ->
    {47,Tlen,Ics,Line,58};
yystate(57, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 4, Tlen);
yystate(57, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 4, Tlen);
yystate(57, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 4, Tlen);
yystate(57, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 4, Tlen);
yystate(57, Ics, Line, Tlen, _, _) ->
    {4,Tlen,Ics,Line,57};
yystate(56, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(56, [50|Ics], Line, Tlen, _, _) ->
    yystate(52, Ics, Line, Tlen+1, 48, Tlen);
yystate(56, [48|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(56, [49|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(56, [C|Ics], Line, Tlen, _, _) when C >= 51, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(56, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(56, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(56, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,56};
yystate(55, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 0, Tlen);
yystate(55, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 0, Tlen);
yystate(55, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 0, Tlen);
yystate(55, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 0, Tlen);
yystate(55, Ics, Line, Tlen, _, _) ->
    {0,Tlen,Ics,Line,55};
yystate(54, Ics, Line, Tlen, _, _) ->
    {50,Tlen,Ics,Line};
yystate(53, [112|Ics], Line, Tlen, _, _) ->
    yystate(49, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, [102|Ics], Line, Tlen, _, _) ->
    yystate(17, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 101 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 103, C =< 111 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, [C|Ics], Line, Tlen, _, _) when C >= 113, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(53, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,53};
yystate(52, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 14, Tlen);
yystate(52, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 14, Tlen);
yystate(52, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 14, Tlen);
yystate(52, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 14, Tlen);
yystate(52, Ics, Line, Tlen, _, _) ->
    {14,Tlen,Ics,Line,52};
yystate(51, [100|Ics], Line, Tlen, _, _) ->
    yystate(55, Ics, Line, Tlen+1, 48, Tlen);
yystate(51, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(51, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(51, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(51, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(51, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(51, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,51};
yystate(50, Ics, Line, Tlen, _, _) ->
    {59,Tlen,Ics,Line};
yystate(49, [114|Ics], Line, Tlen, _, _) ->
    yystate(45, Ics, Line, Tlen+1, 48, Tlen);
yystate(49, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(49, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(49, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(49, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(49, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(49, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,49};
yystate(48, [114|Ics], Line, Tlen, _, _) ->
    yystate(44, Ics, Line, Tlen+1, 48, Tlen);
yystate(48, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(48, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(48, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(48, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(48, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(48, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,48};
yystate(47, [101|Ics], Line, Tlen, _, _) ->
    yystate(51, Ics, Line, Tlen+1, 48, Tlen);
yystate(47, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(47, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(47, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,47};
yystate(46, Ics, Line, Tlen, _, _) ->
    {57,Tlen,Ics,Line};
yystate(45, [101|Ics], Line, Tlen, _, _) ->
    yystate(41, Ics, Line, Tlen+1, 48, Tlen);
yystate(45, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(45, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(45, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,45};
yystate(44, [118|Ics], Line, Tlen, _, _) ->
    yystate(40, Ics, Line, Tlen+1, 48, Tlen);
yystate(44, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(44, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(44, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(44, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 117 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(44, [C|Ics], Line, Tlen, _, _) when C >= 119, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(44, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,44};
yystate(43, [114|Ics], Line, Tlen, _, _) ->
    yystate(47, Ics, Line, Tlen+1, 48, Tlen);
yystate(43, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(43, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(43, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(43, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(43, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(43, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,43};
yystate(42, Ics, Line, Tlen, _, _) ->
    {60,Tlen,Ics,Line};
yystate(41, [99|Ics], Line, Tlen, _, _) ->
    yystate(37, Ics, Line, Tlen+1, 48, Tlen);
yystate(41, [97|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(41, [98|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(41, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(41, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(41, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(41, [C|Ics], Line, Tlen, _, _) when C >= 100, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(41, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,41};
yystate(40, [105|Ics], Line, Tlen, _, _) ->
    yystate(36, Ics, Line, Tlen+1, 48, Tlen);
yystate(40, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(40, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(40, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(40, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(40, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(40, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,40};
yystate(39, [105|Ics], Line, Tlen, _, _) ->
    yystate(43, Ics, Line, Tlen+1, 48, Tlen);
yystate(39, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(39, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(39, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(39, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 104 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(39, [C|Ics], Line, Tlen, _, _) when C >= 106, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(39, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,39};
yystate(38, Ics, Line, Tlen, _, _) ->
    {53,Tlen,Ics,Line};
yystate(37, [97|Ics], Line, Tlen, _, _) ->
    yystate(33, Ics, Line, Tlen+1, 48, Tlen);
yystate(37, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(37, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(37, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(37, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(37, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,37};
yystate(36, [99|Ics], Line, Tlen, _, _) ->
    yystate(32, Ics, Line, Tlen+1, 48, Tlen);
yystate(36, [97|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(36, [98|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(36, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(36, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(36, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(36, [C|Ics], Line, Tlen, _, _) when C >= 100, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(36, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,36};
yystate(35, [117|Ics], Line, Tlen, _, _) ->
    yystate(39, Ics, Line, Tlen+1, 48, Tlen);
yystate(35, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 116 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(35, [C|Ics], Line, Tlen, _, _) when C >= 118, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(35, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,35};
yystate(34, Ics, Line, Tlen, _, _) ->
    {54,Tlen,Ics,Line};
yystate(33, [116|Ics], Line, Tlen, _, _) ->
    yystate(29, Ics, Line, Tlen+1, 48, Tlen);
yystate(33, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(33, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(33, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,33};
yystate(32, [101|Ics], Line, Tlen, _, _) ->
    yystate(28, Ics, Line, Tlen+1, 48, Tlen);
yystate(32, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(32, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(32, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,32};
yystate(31, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 32, Tlen);
yystate(31, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 32, Tlen);
yystate(31, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 32, Tlen);
yystate(31, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 32, Tlen);
yystate(31, Ics, Line, Tlen, _, _) ->
    {32,Tlen,Ics,Line,31};
yystate(30, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 16, Tlen);
yystate(30, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 16, Tlen);
yystate(30, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 16, Tlen);
yystate(30, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 16, Tlen);
yystate(30, Ics, Line, Tlen, _, _) ->
    {16,Tlen,Ics,Line,30};
yystate(29, [101|Ics], Line, Tlen, _, _) ->
    yystate(25, Ics, Line, Tlen+1, 48, Tlen);
yystate(29, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(29, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(29, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,29};
yystate(28, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 33, Tlen);
yystate(28, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 33, Tlen);
yystate(28, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 33, Tlen);
yystate(28, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 33, Tlen);
yystate(28, Ics, Line, Tlen, _, _) ->
    {33,Tlen,Ics,Line,28};
yystate(27, [100|Ics], Line, Tlen, _, _) ->
    yystate(31, Ics, Line, Tlen+1, 48, Tlen);
yystate(27, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(27, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(27, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(27, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(27, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(27, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,27};
yystate(26, [108|Ics], Line, Tlen, _, _) ->
    yystate(30, Ics, Line, Tlen+1, 48, Tlen);
yystate(26, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(26, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(26, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,26};
yystate(25, [100|Ics], Line, Tlen, _, _) ->
    yystate(21, Ics, Line, Tlen+1, 48, Tlen);
yystate(25, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 99 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(25, [C|Ics], Line, Tlen, _, _) when C >= 101, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(25, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,25};
yystate(24, [112|Ics], Line, Tlen, _, _) ->
    yystate(20, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, [101|Ics], Line, Tlen, _, _) ->
    yystate(12, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 111 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, [C|Ics], Line, Tlen, _, _) when C >= 113, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(24, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,24};
yystate(23, [101|Ics], Line, Tlen, _, _) ->
    yystate(27, Ics, Line, Tlen+1, 48, Tlen);
yystate(23, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(23, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(23, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,23};
yystate(22, [111|Ics], Line, Tlen, _, _) ->
    yystate(26, Ics, Line, Tlen+1, 48, Tlen);
yystate(22, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(22, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(22, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,22};
yystate(21, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 27, Tlen);
yystate(21, Ics, Line, Tlen, _, _) ->
    {27,Tlen,Ics,Line,21};
yystate(20, [99|Ics], Line, Tlen, _, _) ->
    yystate(16, Ics, Line, Tlen+1, 48, Tlen);
yystate(20, [97|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(20, [98|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(20, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(20, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(20, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(20, [C|Ics], Line, Tlen, _, _) when C >= 100, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(20, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,20};
yystate(19, [118|Ics], Line, Tlen, _, _) ->
    yystate(23, Ics, Line, Tlen+1, 48, Tlen);
yystate(19, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 117 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(19, [C|Ics], Line, Tlen, _, _) when C >= 119, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(19, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,19};
yystate(18, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 18, Tlen);
yystate(18, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 18, Tlen);
yystate(18, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 18, Tlen);
yystate(18, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 18, Tlen);
yystate(18, Ics, Line, Tlen, _, _) ->
    {18,Tlen,Ics,Line,18};
yystate(17, [97|Ics], Line, Tlen, _, _) ->
    yystate(13, Ics, Line, Tlen+1, 48, Tlen);
yystate(17, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(17, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(17, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(17, [C|Ics], Line, Tlen, _, _) when C >= 98, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(17, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,17};
yystate(16, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 34, Tlen);
yystate(16, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 34, Tlen);
yystate(16, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 34, Tlen);
yystate(16, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 34, Tlen);
yystate(16, Ics, Line, Tlen, _, _) ->
    {34,Tlen,Ics,Line,16};
yystate(15, [114|Ics], Line, Tlen, _, _) ->
    yystate(19, Ics, Line, Tlen+1, 48, Tlen);
yystate(15, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(15, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(15, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(15, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(15, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(15, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,15};
yystate(14, [115|Ics], Line, Tlen, _, _) ->
    yystate(18, Ics, Line, Tlen+1, 48, Tlen);
yystate(14, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(14, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(14, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(14, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(14, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(14, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,14};
yystate(13, [117|Ics], Line, Tlen, _, _) ->
    yystate(9, Ics, Line, Tlen+1, 48, Tlen);
yystate(13, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 116 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(13, [C|Ics], Line, Tlen, _, _) when C >= 118, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(13, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,13};
yystate(12, [116|Ics], Line, Tlen, _, _) ->
    yystate(8, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [115|Ics], Line, Tlen, _, _) ->
    yystate(11, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [114|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [113|Ics], Line, Tlen, _, _) ->
    yystate(35, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [112|Ics], Line, Tlen, _, _) ->
    yystate(59, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 111 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(12, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,12};
yystate(11, [101|Ics], Line, Tlen, _, _) ->
    yystate(15, Ics, Line, Tlen+1, 48, Tlen);
yystate(11, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(11, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(11, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,11};
yystate(10, [101|Ics], Line, Tlen, _, _) ->
    yystate(14, Ics, Line, Tlen+1, 48, Tlen);
yystate(10, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 100 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(10, [C|Ics], Line, Tlen, _, _) when C >= 102, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(10, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,10};
yystate(9, [108|Ics], Line, Tlen, _, _) ->
    yystate(5, Ics, Line, Tlen+1, 48, Tlen);
yystate(9, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 107 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(9, [C|Ics], Line, Tlen, _, _) when C >= 109, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(9, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,9};
yystate(8, [117|Ics], Line, Tlen, _, _) ->
    yystate(4, Ics, Line, Tlen+1, 48, Tlen);
yystate(8, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(8, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(8, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(8, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 116 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(8, [C|Ics], Line, Tlen, _, _) when C >= 118, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(8, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,8};
yystate(7, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 35, Tlen);
yystate(7, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 35, Tlen);
yystate(7, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 35, Tlen);
yystate(7, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 35, Tlen);
yystate(7, Ics, Line, Tlen, _, _) ->
    {35,Tlen,Ics,Line,7};
yystate(6, [116|Ics], Line, Tlen, _, _) ->
    yystate(10, Ics, Line, Tlen+1, 48, Tlen);
yystate(6, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(6, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(6, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,6};
yystate(5, [116|Ics], Line, Tlen, _, _) ->
    yystate(1, Ics, Line, Tlen+1, 48, Tlen);
yystate(5, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(5, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(5, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(5, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 115 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(5, [C|Ics], Line, Tlen, _, _) when C >= 117, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(5, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,5};
yystate(4, [114|Ics], Line, Tlen, _, _) ->
    yystate(0, Ics, Line, Tlen+1, 48, Tlen);
yystate(4, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 113 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(4, [C|Ics], Line, Tlen, _, _) when C >= 115, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(4, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,4};
yystate(3, [115|Ics], Line, Tlen, _, _) ->
    yystate(7, Ics, Line, Tlen+1, 48, Tlen);
yystate(3, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(3, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(3, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(3, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 114 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(3, [C|Ics], Line, Tlen, _, _) when C >= 116, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(3, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,3};
yystate(2, [122|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, [121|Ics], Line, Tlen, _, _) ->
    yystate(6, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, [111|Ics], Line, Tlen, _, _) ->
    yystate(22, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 110 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, [C|Ics], Line, Tlen, _, _) when C >= 112, C =< 120 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(2, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,2};
yystate(1, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 25, Tlen);
yystate(1, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 25, Tlen);
yystate(1, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 25, Tlen);
yystate(1, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 25, Tlen);
yystate(1, Ics, Line, Tlen, _, _) ->
    {25,Tlen,Ics,Line,1};
yystate(0, [110|Ics], Line, Tlen, _, _) ->
    yystate(3, Ics, Line, Tlen+1, 48, Tlen);
yystate(0, [95|Ics], Line, Tlen, _, _) ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 48, C =< 57 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 65, C =< 90 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 97, C =< 109 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(0, [C|Ics], Line, Tlen, _, _) when C >= 111, C =< 122 ->
    yystate(180, Ics, Line, Tlen+1, 48, Tlen);
yystate(0, Ics, Line, Tlen, _, _) ->
    {48,Tlen,Ics,Line,0};
yystate(S, Ics, Line, Tlen, Action, Alen) ->
    {Action,Alen,Tlen,Ics,Line,S}.

%% yyaction(Action, TokenLength, TokenChars, TokenLine) ->
%% {token,Token} | {end_token, Token} | skip_token | {error,String}.
%% Generated action function.

yyaction(0, _, _, TokenLine) ->
    yyaction_0(TokenLine);
yyaction(1, _, _, TokenLine) ->
    yyaction_1(TokenLine);
yyaction(2, _, _, TokenLine) ->
    yyaction_2(TokenLine);
yyaction(3, _, _, TokenLine) ->
    yyaction_3(TokenLine);
yyaction(4, _, _, TokenLine) ->
    yyaction_4(TokenLine);
yyaction(5, _, _, TokenLine) ->
    yyaction_5(TokenLine);
yyaction(6, _, _, TokenLine) ->
    yyaction_6(TokenLine);
yyaction(7, _, _, TokenLine) ->
    yyaction_7(TokenLine);
yyaction(8, _, _, TokenLine) ->
    yyaction_8(TokenLine);
yyaction(9, _, _, TokenLine) ->
    yyaction_9(TokenLine);
yyaction(10, _, _, TokenLine) ->
    yyaction_10(TokenLine);
yyaction(11, _, _, TokenLine) ->
    yyaction_11(TokenLine);
yyaction(12, _, _, TokenLine) ->
    yyaction_12(TokenLine);
yyaction(13, _, _, TokenLine) ->
    yyaction_13(TokenLine);
yyaction(14, _, _, TokenLine) ->
    yyaction_14(TokenLine);
yyaction(15, _, _, TokenLine) ->
    yyaction_15(TokenLine);
yyaction(16, _, _, TokenLine) ->
    yyaction_16(TokenLine);
yyaction(17, _, _, TokenLine) ->
    yyaction_17(TokenLine);
yyaction(18, _, _, TokenLine) ->
    yyaction_18(TokenLine);
yyaction(19, _, _, TokenLine) ->
    yyaction_19(TokenLine);
yyaction(20, _, _, TokenLine) ->
    yyaction_20(TokenLine);
yyaction(21, _, _, TokenLine) ->
    yyaction_21(TokenLine);
yyaction(22, _, _, TokenLine) ->
    yyaction_22(TokenLine);
yyaction(23, _, _, TokenLine) ->
    yyaction_23(TokenLine);
yyaction(24, _, _, TokenLine) ->
    yyaction_24(TokenLine);
yyaction(25, _, _, TokenLine) ->
    yyaction_25(TokenLine);
yyaction(26, _, _, TokenLine) ->
    yyaction_26(TokenLine);
yyaction(27, _, _, TokenLine) ->
    yyaction_27(TokenLine);
yyaction(28, _, _, TokenLine) ->
    yyaction_28(TokenLine);
yyaction(29, _, _, TokenLine) ->
    yyaction_29(TokenLine);
yyaction(30, _, _, TokenLine) ->
    yyaction_30(TokenLine);
yyaction(31, _, _, TokenLine) ->
    yyaction_31(TokenLine);
yyaction(32, _, _, TokenLine) ->
    yyaction_32(TokenLine);
yyaction(33, _, _, TokenLine) ->
    yyaction_33(TokenLine);
yyaction(34, _, _, TokenLine) ->
    yyaction_34(TokenLine);
yyaction(35, _, _, TokenLine) ->
    yyaction_35(TokenLine);
yyaction(36, _, _, TokenLine) ->
    yyaction_36(TokenLine);
yyaction(37, _, _, TokenLine) ->
    yyaction_37(TokenLine);
yyaction(38, _, _, TokenLine) ->
    yyaction_38(TokenLine);
yyaction(39, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_39(TokenChars, TokenLine);
yyaction(40, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_40(TokenChars, TokenLine);
yyaction(41, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_41(TokenChars, TokenLine);
yyaction(42, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_42(TokenChars, TokenLine);
yyaction(43, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_43(TokenChars, TokenLine);
yyaction(44, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_44(TokenChars, TokenLine);
yyaction(45, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_45(TokenChars, TokenLine);
yyaction(46, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_46(TokenChars, TokenLine);
yyaction(47, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_47(TokenChars, TokenLine);
yyaction(48, TokenLen, YYtcs, TokenLine) ->
    TokenChars = yypre(YYtcs, TokenLen),
    yyaction_48(TokenChars, TokenLine);
yyaction(49, _, _, TokenLine) ->
    yyaction_49(TokenLine);
yyaction(50, _, _, TokenLine) ->
    yyaction_50(TokenLine);
yyaction(51, _, _, TokenLine) ->
    yyaction_51(TokenLine);
yyaction(52, _, _, TokenLine) ->
    yyaction_52(TokenLine);
yyaction(53, _, _, TokenLine) ->
    yyaction_53(TokenLine);
yyaction(54, _, _, TokenLine) ->
    yyaction_54(TokenLine);
yyaction(55, _, _, TokenLine) ->
    yyaction_55(TokenLine);
yyaction(56, _, _, TokenLine) ->
    yyaction_56(TokenLine);
yyaction(57, _, _, TokenLine) ->
    yyaction_57(TokenLine);
yyaction(58, _, _, TokenLine) ->
    yyaction_58(TokenLine);
yyaction(59, _, _, TokenLine) ->
    yyaction_59(TokenLine);
yyaction(60, _, _, TokenLine) ->
    yyaction_60(TokenLine);
yyaction(61, _, _, _) ->
    yyaction_61();
yyaction(62, _, _, _) ->
    yyaction_62();
yyaction(63, _, _, _) ->
    yyaction_63();
yyaction(64, _, _, _) ->
    yyaction_64();
yyaction(_, _, _, _) -> error.

-compile({inline,yyaction_0/1}).
-file("src/gpb_scan.xrl", 28).
yyaction_0(TokenLine) ->
     { token, { required, TokenLine } } .

-compile({inline,yyaction_1/1}).
-file("src/gpb_scan.xrl", 29).
yyaction_1(TokenLine) ->
     { token, { optional, TokenLine } } .

-compile({inline,yyaction_2/1}).
-file("src/gpb_scan.xrl", 30).
yyaction_2(TokenLine) ->
     { token, { repeated, TokenLine } } .

-compile({inline,yyaction_3/1}).
-file("src/gpb_scan.xrl", 31).
yyaction_3(TokenLine) ->
     { token, { oneof, TokenLine } } .

-compile({inline,yyaction_4/1}).
-file("src/gpb_scan.xrl", 33).
yyaction_4(TokenLine) ->
     { token, { double, TokenLine } } .

-compile({inline,yyaction_5/1}).
-file("src/gpb_scan.xrl", 34).
yyaction_5(TokenLine) ->
     { token, { float, TokenLine } } .

-compile({inline,yyaction_6/1}).
-file("src/gpb_scan.xrl", 35).
yyaction_6(TokenLine) ->
     { token, { int32, TokenLine } } .

-compile({inline,yyaction_7/1}).
-file("src/gpb_scan.xrl", 36).
yyaction_7(TokenLine) ->
     { token, { int64, TokenLine } } .

-compile({inline,yyaction_8/1}).
-file("src/gpb_scan.xrl", 37).
yyaction_8(TokenLine) ->
     { token, { uint32, TokenLine } } .

-compile({inline,yyaction_9/1}).
-file("src/gpb_scan.xrl", 38).
yyaction_9(TokenLine) ->
     { token, { uint64, TokenLine } } .

-compile({inline,yyaction_10/1}).
-file("src/gpb_scan.xrl", 39).
yyaction_10(TokenLine) ->
     { token, { sint32, TokenLine } } .

-compile({inline,yyaction_11/1}).
-file("src/gpb_scan.xrl", 40).
yyaction_11(TokenLine) ->
     { token, { sint64, TokenLine } } .

-compile({inline,yyaction_12/1}).
-file("src/gpb_scan.xrl", 41).
yyaction_12(TokenLine) ->
     { token, { fixed32, TokenLine } } .

-compile({inline,yyaction_13/1}).
-file("src/gpb_scan.xrl", 42).
yyaction_13(TokenLine) ->
     { token, { fixed64, TokenLine } } .

-compile({inline,yyaction_14/1}).
-file("src/gpb_scan.xrl", 43).
yyaction_14(TokenLine) ->
     { token, { sfixed32, TokenLine } } .

-compile({inline,yyaction_15/1}).
-file("src/gpb_scan.xrl", 44).
yyaction_15(TokenLine) ->
     { token, { sfixed64, TokenLine } } .

-compile({inline,yyaction_16/1}).
-file("src/gpb_scan.xrl", 45).
yyaction_16(TokenLine) ->
     { token, { bool, TokenLine } } .

-compile({inline,yyaction_17/1}).
-file("src/gpb_scan.xrl", 46).
yyaction_17(TokenLine) ->
     { token, { string, TokenLine } } .

-compile({inline,yyaction_18/1}).
-file("src/gpb_scan.xrl", 47).
yyaction_18(TokenLine) ->
     { token, { bytes, TokenLine } } .

-compile({inline,yyaction_19/1}).
-file("src/gpb_scan.xrl", 48).
yyaction_19(TokenLine) ->
     { token, { map, TokenLine } } .

-compile({inline,yyaction_20/1}).
-file("src/gpb_scan.xrl", 50).
yyaction_20(TokenLine) ->
     { token, { package, TokenLine } } .

-compile({inline,yyaction_21/1}).
-file("src/gpb_scan.xrl", 51).
yyaction_21(TokenLine) ->
     { token, { message, TokenLine } } .

-compile({inline,yyaction_22/1}).
-file("src/gpb_scan.xrl", 52).
yyaction_22(TokenLine) ->
     { token, { enum, TokenLine } } .

-compile({inline,yyaction_23/1}).
-file("src/gpb_scan.xrl", 53).
yyaction_23(TokenLine) ->
     { token, { option, TokenLine } } .

-compile({inline,yyaction_24/1}).
-file("src/gpb_scan.xrl", 54).
yyaction_24(TokenLine) ->
     { token, { import, TokenLine } } .

-compile({inline,yyaction_25/1}).
-file("src/gpb_scan.xrl", 55).
yyaction_25(TokenLine) ->
     { token, { default, TokenLine } } .

-compile({inline,yyaction_26/1}).
-file("src/gpb_scan.xrl", 56).
yyaction_26(TokenLine) ->
     { token, { packed, TokenLine } } .

-compile({inline,yyaction_27/1}).
-file("src/gpb_scan.xrl", 57).
yyaction_27(TokenLine) ->
     { token, { deprecated, TokenLine } } .

-compile({inline,yyaction_28/1}).
-file("src/gpb_scan.xrl", 58).
yyaction_28(TokenLine) ->
     { token, { extensions, TokenLine } } .

-compile({inline,yyaction_29/1}).
-file("src/gpb_scan.xrl", 59).
yyaction_29(TokenLine) ->
     { token, { extend, TokenLine } } .

-compile({inline,yyaction_30/1}).
-file("src/gpb_scan.xrl", 60).
yyaction_30(TokenLine) ->
     { token, { to, TokenLine } } .

-compile({inline,yyaction_31/1}).
-file("src/gpb_scan.xrl", 61).
yyaction_31(TokenLine) ->
     { token, { max, TokenLine } } .

-compile({inline,yyaction_32/1}).
-file("src/gpb_scan.xrl", 62).
yyaction_32(TokenLine) ->
     { token, { reserved, TokenLine } } .

-compile({inline,yyaction_33/1}).
-file("src/gpb_scan.xrl", 64).
yyaction_33(TokenLine) ->
     { token, { service, TokenLine } } .

-compile({inline,yyaction_34/1}).
-file("src/gpb_scan.xrl", 65).
yyaction_34(TokenLine) ->
     { token, { rpc, TokenLine } } .

-compile({inline,yyaction_35/1}).
-file("src/gpb_scan.xrl", 66).
yyaction_35(TokenLine) ->
     { token, { returns, TokenLine } } .

-compile({inline,yyaction_36/1}).
-file("src/gpb_scan.xrl", 68).
yyaction_36(TokenLine) ->
     { token, { bool_lit, TokenLine, true } } .

-compile({inline,yyaction_37/1}).
-file("src/gpb_scan.xrl", 69).
yyaction_37(TokenLine) ->
     { token, { bool_lit, TokenLine, false } } .

-compile({inline,yyaction_38/1}).
-file("src/gpb_scan.xrl", 71).
yyaction_38(TokenLine) ->
     { token, { syntax, TokenLine } } .

-compile({inline,yyaction_39/2}).
-file("src/gpb_scan.xrl", 75).
yyaction_39(TokenChars, TokenLine) ->
     { token, { str_lit, TokenLine, string_value (TokenChars) } } .

-compile({inline,yyaction_40/2}).
-file("src/gpb_scan.xrl", 77).
yyaction_40(TokenChars, TokenLine) ->
     { token, { str_lit, TokenLine, string_value (TokenChars) } } .

-compile({inline,yyaction_41/2}).
-file("src/gpb_scan.xrl", 80).
yyaction_41(TokenChars, TokenLine) ->
     { token, { float_lit, TokenLine, str_to_float_1 (TokenChars) } } .

-compile({inline,yyaction_42/2}).
-file("src/gpb_scan.xrl", 82).
yyaction_42(TokenChars, TokenLine) ->
     { token, { float_lit, TokenLine, str_to_float_2 (TokenChars) } } .

-compile({inline,yyaction_43/2}).
-file("src/gpb_scan.xrl", 84).
yyaction_43(TokenChars, TokenLine) ->
     { token, { float_lit, TokenLine, str_to_float_3 (TokenChars) } } .

-compile({inline,yyaction_44/2}).
-file("src/gpb_scan.xrl", 86).
yyaction_44(TokenChars, TokenLine) ->
     { token, { float_lit, TokenLine, str_to_float_4 (TokenChars) } } .

-compile({inline,yyaction_45/2}).
-file("src/gpb_scan.xrl", 92).
yyaction_45(TokenChars, TokenLine) ->
     { token, { hex_lit, TokenLine, hexstr_to_integer (TokenChars) } } .

-compile({inline,yyaction_46/2}).
-file("src/gpb_scan.xrl", 93).
yyaction_46(TokenChars, TokenLine) ->
     { token, { oct_lit, TokenLine, octstr_to_integer (TokenChars) } } .

-compile({inline,yyaction_47/2}).
-file("src/gpb_scan.xrl", 94).
yyaction_47(TokenChars, TokenLine) ->
     { token, { dec_lit, TokenLine, list_to_integer (TokenChars) } } .

-compile({inline,yyaction_48/2}).
-file("src/gpb_scan.xrl", 96).
yyaction_48(TokenChars, TokenLine) ->
     { token, { identifier, TokenLine, TokenChars } } .

-compile({inline,yyaction_49/1}).
-file("src/gpb_scan.xrl", 98).
yyaction_49(TokenLine) ->
     { token, { '.', TokenLine } } .

-compile({inline,yyaction_50/1}).
-file("src/gpb_scan.xrl", 99).
yyaction_50(TokenLine) ->
     { token, { ';', TokenLine } } .

-compile({inline,yyaction_51/1}).
-file("src/gpb_scan.xrl", 100).
yyaction_51(TokenLine) ->
     { token, { '{', TokenLine } } .

-compile({inline,yyaction_52/1}).
-file("src/gpb_scan.xrl", 101).
yyaction_52(TokenLine) ->
     { token, { '}', TokenLine } } .

-compile({inline,yyaction_53/1}).
-file("src/gpb_scan.xrl", 102).
yyaction_53(TokenLine) ->
     { token, { '[', TokenLine } } .

-compile({inline,yyaction_54/1}).
-file("src/gpb_scan.xrl", 103).
yyaction_54(TokenLine) ->
     { token, { ']', TokenLine } } .

-compile({inline,yyaction_55/1}).
-file("src/gpb_scan.xrl", 104).
yyaction_55(TokenLine) ->
     { token, { '(', TokenLine } } .

-compile({inline,yyaction_56/1}).
-file("src/gpb_scan.xrl", 105).
yyaction_56(TokenLine) ->
     { token, { ')', TokenLine } } .

-compile({inline,yyaction_57/1}).
-file("src/gpb_scan.xrl", 106).
yyaction_57(TokenLine) ->
     { token, { '=', TokenLine } } .

-compile({inline,yyaction_58/1}).
-file("src/gpb_scan.xrl", 107).
yyaction_58(TokenLine) ->
     { token, { ',', TokenLine } } .

-compile({inline,yyaction_59/1}).
-file("src/gpb_scan.xrl", 108).
yyaction_59(TokenLine) ->
     { token, { '<', TokenLine } } .

-compile({inline,yyaction_60/1}).
-file("src/gpb_scan.xrl", 109).
yyaction_60(TokenLine) ->
     { token, { '>', TokenLine } } .

-compile({inline,yyaction_61/0}).
-file("src/gpb_scan.xrl", 112).
yyaction_61() ->
     skip_token .

-compile({inline,yyaction_62/0}).
-file("src/gpb_scan.xrl", 113).
yyaction_62() ->
     skip_token .

-compile({inline,yyaction_63/0}).
-file("src/gpb_scan.xrl", 114).
yyaction_63() ->
     skip_token .

-compile({inline,yyaction_64/0}).
-file("src/gpb_scan.xrl", 115).
yyaction_64() ->
     skip_token .

-file("/usr/local/lib/erlang/lib/parsetools-2.1.1/include/leexinc.hrl", 290).
